unit Crt;

{*******************************************************}
{                                                       }
{       Windows CRT Interface Unit                      }
{                                                       }
{       WinCrt original interface                       }
{       Copytight (C) 1991,95 Borland International     }
{                                                       }
{       Win32 Console compatible and additional         }
{       2000..2003 YT                                   }
{                                                       }
{*******************************************************}

{$WARN SYMBOL_PLATFORM OFF}

interface

uses
	Windows;

{ WinCrt compatible }

const
	BLACK        = 0;
	BLUE         = FOREGROUND_BLUE;
	GREEN        = FOREGROUND_GREEN;
	CYAN         = FOREGROUND_GREEN or FOREGROUND_BLUE;
	RED          = FOREGROUND_RED;
	MAGENTA      = FOREGROUND_RED or FOREGROUND_BLUE;
	BROWN        = FOREGROUND_RED or FOREGROUND_GREEN;
	LIGHTGRAY    = FOREGROUND_RED or FOREGROUND_GREEN or FOREGROUND_BLUE;
	DARKGRAY     = 8;
	LIGHTBLUE    = 9;
	LIGHTGREEN   = 10;
	LIGHTCYAN    = 11;
	LIGHTRED     = 12;
	LIGHTMAGENTA = 13;
	YELLOW       = 14;
	WHITE        = 15;

procedure TextColor(Color: Cardinal);
procedure TextBackground(Color: Cardinal);
procedure LowVideo;
procedure HighVideo;
procedure NormVideo;
procedure GotoXY(X, Y: Integer);
procedure ClrScr;
procedure ClrEol;
function WhereX: Integer;
function WhereY: Integer;
procedure WriteBuf(Buffer: PChar; Count: Word);
procedure WriteChar(Ch: Char);
function ReadBuf(Buffer: PChar; Count: Word): Word;
function ReadKey: Char;
function KeyPressed: Boolean;

procedure AssignCrt(var F: Text);
procedure AssignErr(var F: Text);

{ Win32 Crt }

function IsCrt: Boolean;
function IsForeground: Boolean;
procedure CrtSetTitle(const Title: string);
function CrtTitle: string;

procedure CrtSetSize(const Size: TSize); overload;
procedure CrtSetSize(var F: Text; const Size: TSize); overload;
procedure CrtSetSize(Screen: THandle; const Size: TSize); overload;
function CrtSize: TSize; overload;
function CrtSize(var F: Text): TSize; overload;
function CrtSize(Screen: THandle): TSize; overload;
procedure CrtSetTextAttribute(Attribute: Cardinal); overload;
procedure CrtSetTextAttribute(var F: Text; Attribute: Cardinal); overload;
procedure CrtSetTextAttribute(Screen: THandle; Attribute: Cardinal); overload;
function CrtTextAttribute: Cardinal; overload;
function CrtTextAttribute(var F: Text): Cardinal; overload;
function CrtTextAttribute(Screen: THandle): Cardinal; overload;
procedure CrtLocate(X, Y: Integer); overload;
procedure CrtLocate(var F: Text; X, Y: Integer); overload;
procedure CrtLocate(Screen: THandle; X, Y: Integer); overload;
function CrtCursorPosition: TPoint; overload;
function CrtCursorPosition(var F: Text): TPoint; overload;
function CrtCursorPosition(Screen: THandle): TPoint; overload;
procedure CrtShowCursor(Visible: Boolean); overload;
procedure CrtShowCursor(var F: Text; Visible: Boolean); overload;
procedure CrtShowCursor(Screen: THandle; Visible: Boolean); overload;
function CrtCursorVisible: Boolean; overload;
function CrtCursorVisible(var F: Text): Boolean; overload;
function CrtCursorVisible(Screen: THandle): Boolean; overload;
procedure CrtClear; overload;
procedure CrtClear(var F: Text); overload;
procedure CrtClear(Screen: THandle); overload;
procedure CrtClearEoln; overload;
procedure CrtClearEoln(var F: Text); overload;
procedure CrtClearEoln(Screen: THandle); overload;

function CrtWrite(Screen: THandle; const Text: string): Cardinal; overload;
function CrtWrite(Screen: THandle; T: PChar; L: Cardinal): Cardinal; overload;
function CrtRead(B: PChar; L: Cardinal): Cardinal; overload;
function CrtRead(var F: Text; B: PChar; L: Cardinal): Cardinal; overload;
function CrtRead(Input: THandle; B: PChar; L: Cardinal): Cardinal; overload;
function CrtReadChar: Char; overload;
function CrtReadChar(Input: THandle): Char; overload;
function CrtPeekChar: Char; overload;
function CrtPeekChar(Input: THandle): Char; overload;

function NewScreen: THandle;
procedure AssignNewScreen(var F: Text);
procedure ActivateScreen(var F: Text);

function CrtHasEvents(var F: Text): Boolean; overload;
function CrtHasEvents(Input: THandle): Boolean; overload;
function CrtGetEvent(var F: Text; var Event: INPUT_RECORD): Boolean; overload;
function CrtGetEvent(Input: THandle; var Event: INPUT_RECORD): Boolean; overload;

{ Direct Access }

procedure CrtDirectWrite(var F: Text; X, Y: Integer; const Text: String; Attribute: Cardinal); overload;
procedure CrtDirectWrite(Screen: THandle; X, Y: Integer; const Text: String; Attribute: Cardinal); overload;
procedure CrtDirectWrite(Screen: THandle; X, Y: Integer; P: PChar; L: Cardinal; Attribute: Cardinal); overload;

{ Text Mode Graphics }

procedure CrtLine(var F: Text; X1, Y1, X2, Y2: Integer; const S: string; Attribute: Cardinal); overload;
procedure CrtLine(Screen: THandle; X1, Y1, X2, Y2: Integer; const S: string; Attribute: Cardinal); overload;
procedure CrtLine(var F: Text; X1, Y1, X2, Y2: Integer; const S: Char; Attribute: Cardinal); overload;
procedure CrtLine(Screen: THandle; X1, Y1, X2, Y2: Integer; const S: Char; Attribute: Cardinal); overload;

{}

procedure CrtWriteAttribute(Screen: THandle; const R: TRect; Attribute: Cardinal);
procedure CrtWriteDirectTransparent(Screen: THandle; X, Y: Integer; const Text: String; ForeAttribute: Cardinal);

const
	ZeroCoord: COORD = (X: 0; Y: 0);
	RunUnderWin32Console = 'This program must be run under Win32 Console.';

{ API }

function GetConsoleWindow: THandle; stdcall; external 'kernel32.dll';

implementation

{ WinCrt compatible }

procedure TextColor(Color: Cardinal);
begin
	CrtSetTextAttribute(Color or (CrtTextAttribute and not $f));
end;

procedure TextBackground(Color: Cardinal);
begin
	CrtSetTextAttribute((Color shl 4) or (CrtTextAttribute and not $f0));
end;

procedure LowVideo;
begin
	CrtSetTextAttribute($07)
end;

procedure HighVideo;
begin
	CrtSetTextAttribute($8f)
end;

procedure NormVideo;
begin
	CrtSetTextAttribute(7)
end;

procedure GotoXY(X, Y: Integer);
begin
	CrtLocate(X - 1, Y - 1);
end;

function WhereX: Integer;
begin
	Result := CrtCursorPosition.X + 1;
end;

function WhereY: Integer;
begin
	Result := CrtCursorPosition.Y + 1;
end;

procedure ClrScr;
begin
	CrtClear;
end;

procedure ClrEol;
begin
	CrtClearEoln;
end;

procedure WriteBuf(Buffer: PChar; Count: Word);
begin
	CrtWrite(GetStdHandle(STD_OUTPUT_HANDLE), Buffer, Count)
end;

procedure WriteChar(Ch: Char);
begin
	CrtWrite(GetStdHandle(STD_OUTPUT_HANDLE), @Ch, 1)
end;

function ReadBuf(Buffer: PChar; Count: Word): Word;
begin
	Result := CrtRead(Buffer, Count);
end;

function ReadKey: Char;
begin
	Result := CrtReadChar;
end;

function KeyPressed: Boolean;
begin
	Result := CrtPeekChar <> #0;
end;

procedure AssignCrt(var F: Text);
begin
	Assign(F, '');
end;

procedure AssignErr(var F: Text);

	function NoAction(var T: Text): Integer;
	begin
		Result := 0;
	end;

begin
	Assign(F, '');
	Rewrite(F);
	TTextRec(F).Handle := GetStdHandle(STD_ERROR_HANDLE);
	TTextRec(F).OpenFunc := @NoAction;
end;

{ Win32 Crt }

function IsCrt: Boolean;
var
	Mode: DWORD;
	Info: CONSOLE_SCREEN_BUFFER_INFO;
begin
	Result:= GetConsoleMode(GetStdHandle(STD_INPUT_HANDLE), Mode) and
					GetConsoleScreenBufferInfo(GetStdHandle(STD_OUTPUT_HANDLE), Info);
end;

function IsForeground: Boolean;
begin
	Result := GetConsoleWindow = GetForegroundWindow
end;

procedure CrtSetTitle(const Title: string);
begin
	SetConsoleTitle(PChar(Title))
end;

function CrtTitle: string;
begin
	SetLength(Result, 1024);
	SetLength(Result, GetConsoleTitle(PChar(Result), Length(Result)) - 1)
end;	

procedure CrtSetSize(const Size: TSize); overload;
begin
	CrtSetSize(GetStdHandle(STD_OUTPUT_HANDLE), Size)
end;

procedure CrtSetSize(var F: Text; const Size: TSize); overload;
begin
  CrtSetSize(TTextRec(F).Handle, Size)
end;

procedure CrtSetSize(Screen: THandle; const Size: TSize); overload;
var
	T: COORD;
	Info: CONSOLE_SCREEN_BUFFER_INFO;
	R: SMALL_RECT;
begin
	T.X := Size.cx;
	T.Y := Size.cy;
	SetConsoleScreenBufferSize(Screen, T);
	GetConsoleScreenBufferInfo(Screen, Info);
	R.Left := 0;
	R.Top := 0;
	R.Right := Info.dwMaximumWindowSize.X - 1;
	R.Bottom := Info.dwMaximumWindowSize.Y - 1;
	SetConsoleWindowInfo(Screen, True, R);
end;

function CrtSize: TSize; overload;
begin
	Result := CrtSize(GetStdHandle(STD_OUTPUT_HANDLE));
end;

function CrtSize(var F: Text): TSize; overload;
begin
	Result := CrtSize(TTextRec(F).Handle);
end;

function CrtSize(Screen: THandle): TSize; overload;
var
	Info: CONSOLE_SCREEN_BUFFER_INFO;
begin
	GetConsoleScreenBufferInfo(Screen, Info);
	Result.CX := Info.dwMaximumWindowSize.X;
	Result.CY := Info.dwMaximumWindowSize.Y;
end;

procedure CrtSetTextAttribute(Attribute: Cardinal); overload;
begin
	CrtSetTextAttribute(GetStdHandle(STD_OUTPUT_HANDLE), Attribute);
end;

procedure CrtSetTextAttribute(var F: Text; Attribute: Cardinal); overload;
begin
	CrtSetTextAttribute(TTextRec(F).Handle, Attribute);
end;

procedure CrtSetTextAttribute(Screen: THandle; Attribute: Cardinal); overload;
begin
	SetConsoleTextAttribute(Screen, Attribute);
end;

function CrtTextAttribute: Cardinal; overload;
begin
	Result := CrtTextAttribute(GetStdHandle(STD_OUTPUT_HANDLE))
end;

function CrtTextAttribute(var F: Text): Cardinal; overload;
begin
	Result := CrtTextAttribute(TTextRec(F).Handle);
end;

function CrtTextAttribute(Screen: THandle): Cardinal; overload;
var
	Info: CONSOLE_SCREEN_BUFFER_INFO;
begin
	GetConsoleScreenBufferInfo(Screen, Info);
	Result := Info.wAttributes
end;

procedure CrtLocate(X, Y: Integer); overload;
begin
	CrtLocate(GetStdHandle(STD_OUTPUT_HANDLE), X, Y)
end;

procedure CrtLocate(var F: Text; X, Y: Integer); overload;
begin
	CrtLocate(TTextRec(F).Handle, X, Y);
end;

procedure CrtLocate(Screen: THandle; X, Y: Integer); overload;
var
	P: COORD;
begin
	P.X:= X;
	P.Y:= Y;
	SetConsoleCursorPosition(Screen, P);
end;

function CrtCursorPosition: TPoint; overload;
begin
	Result := CrtCursorPosition(GetStdHandle(STD_OUTPUT_HANDLE));
end;

function CrtCursorPosition(var F: Text): TPoint; overload;
begin
	Result := CrtCursorPosition(TTextRec(F).Handle);
end;

function CrtCursorPosition(Screen: THandle): TPoint; overload;
var
	Info: CONSOLE_SCREEN_BUFFER_INFO;
begin
	GetConsoleScreenBufferInfo(Screen, Info);
	Result.X := Info.dwCursorPosition.X;
	Result.Y := Info.dwCursorPosition.Y;
end;

procedure CrtShowCursor(Visible: Boolean); overload;
begin
	CrtShowCursor(GetStdHandle(STD_OUTPUT_HANDLE), Visible);
end;

procedure CrtShowCursor(var F: Text; Visible: Boolean); overload;
begin
	CrtShowCursor(TTextRec(F).Handle, Visible);
end;

procedure CrtShowCursor(Screen: THandle; Visible: Boolean); overload;
var
	Info: CONSOLE_CURSOR_INFO;
begin
	GetConsoleCursorInfo(Screen, Info);
	Info.bVisible := Visible;
	SetConsoleCursorInfo(Screen, Info);
end;

function CrtCursorVisible: Boolean; overload;
begin
	Result := CrtCursorVisible(GetStdHandle(STD_OUTPUT_HANDLE))
end;

function CrtCursorVisible(var F: Text): Boolean; overload;
begin
	Result := CrtCursorVisible(TTextRec(F).Handle)
end;

function CrtCursorVisible(Screen: THandle): Boolean; overload;
var
	Info: CONSOLE_CURSOR_INFO;
begin
	GetConsoleCursorInfo(Screen, Info);
	Result := Info.bVisible
end;

procedure CrtClear; overload;
begin
	CrtClear(GetStdHandle(STD_OUTPUT_HANDLE));
end;

procedure CrtClear(var F: Text); overload;
begin
	CrtClear(TTextRec(F).Handle);
end;

procedure CrtClear(Screen: THandle); overload;
const
	BufferSize: COORD = (X: $100; Y: $100);
var
	Buffer: array[0..$ff, 0..$ff] of CHAR_INFO;
	R: SMALL_RECT;
	S: TSize;
begin
	FillChar(Buffer, SizeOf(Buffer), 0);
	S := CrtSize(Screen);
	R.Left := 0; R.Top := 0; R.Right := S.CX - 1; R.Bottom := S.CY - 1;
	WriteConsoleOutput(Screen, @Buffer, BufferSize, ZeroCoord, R);
	CrtLocate(Screen, 0, 0);
end;

procedure CrtClearEoln; overload;
begin
	CrtClearEoln(GetStdHandle(STD_OUTPUT_HANDLE))
end;

procedure CrtClearEoln(var F: Text); overload;
begin
	CrtClearEoln(TTextRec(F).Handle);
end;

procedure CrtClearEoln(Screen: THandle); overload;
const
	BufferSize: COORD = (X: $100; Y: 1);
var
	Buffer: array[0..$ff, 0..0] of CHAR_INFO;
	R: SMALL_RECT;
	P: TPoint;
	S: TSize;
begin
	FillChar(Buffer, SizeOf(Buffer), 0);
	P := CrtCursorPosition(Screen);
	S := CrtSize(Screen);
	R.Left := P.X; R.Top := P.Y; R.Right := S.CX - 1; R.Bottom := P.Y;
	WriteConsoleOutput(Screen, @Buffer, BufferSize, ZeroCoord, R);
end;

function CrtWrite(Screen: THandle; const Text: String): Cardinal;
begin
	WriteConsole(Screen, PChar(Text), Length(Text), Result, nil);
end;

function CrtWrite(Screen: THandle; T: PChar; L: Cardinal): Cardinal;
begin
	WriteConsole(Screen, T, L, Result, nil);
end;

function CrtRead(B: PChar; L: Cardinal): Cardinal; overload;
begin
	Result := CrtRead(GetStdHandle(STD_INPUT_HANDLE), B, L);
end;

function CrtRead(var F: Text; B: PChar; L: Cardinal): Cardinal; overload;
begin
	Result := CrtRead(TTextRec(F).Handle, B, L);
end;

function CrtRead(Input: THandle; B: PChar; L: Cardinal): Cardinal; overload;
begin
	ReadConsole(Input, B, L, Result, nil);
end;

function CrtReadChar: Char; overload;
begin
	Result := CrtReadChar(GetStdHandle(STD_INPUT_HANDLE));
end;

function CrtReadChar(Input: THandle): Char; overload;
var
	OrgMode, R: DWORD;
begin
	GetConsoleMode(Input, OrgMode);
	SetConsoleMode(Input, OrgMode and
					not (ENABLE_LINE_INPUT or ENABLE_ECHO_INPUT));
	ReadConsole(Input, @Result, 1, R, nil);
	SetConsoleMode(Input, OrgMode);
end;

function CrtPeekChar: Char; overload;
begin
	Result := CrtPeekChar(GetStdHandle(STD_INPUT_HANDLE));
end;

function CrtPeekChar(Input: THandle): Char; overload;
type
	TEventArray = array[0..32767] of INPUT_RECORD;
var
	PEvents: ^TEventArray;
	I, Count, Readed: Cardinal;
begin
	Result := #0;
	GetNumberOfConsoleInputEvents(Input, Count);
	if Count > 0 then
	begin
		GetMem(PEvents, SizeOf(INPUT_RECORD) * Count);
		try
			PeekConsoleInput(Input, PEvents^[0], Count, Readed);
			for I := 0 to Readed - 1 do
				if PEvents^[I].EventType = KEY_EVENT then
					if PEvents^[I].Event.KeyEvent.bKeyDown then
						if PEvents^[I].Event.KeyEvent.AsciiChar <> #0 then
						begin
							Result := PEvents^[I].Event.KeyEvent.AsciiChar;
							Exit;
						end;
		finally
			FreeMem(PEvents);
		end;
	end;
end;

function NewScreen: THandle;
begin
	Result := CreateConsoleScreenBuffer(
					GENERIC_READ or GENERIC_WRITE, 0, nil, CONSOLE_TEXTMODE_BUFFER, nil)
end;

procedure AssignNewScreen(var F: Text);

	function NoAction(var T: Text): Integer;
	begin
		Result := 0;
	end;

	function OnClose(var T: Text): Integer;
	begin
		CloseHandle(TTextRec(T).Handle);
		Result := 0;
	end;

begin
	Assign(F, '');
	Rewrite(F);
	TTextRec(F).Handle := NewScreen;
	TTextRec(F).OpenFunc := @NoAction;
	TTextRec(F).CloseFunc := @OnClose;
end;

procedure ActivateScreen(var F: Text);
begin
	SetConsoleActiveScreenBuffer(TTextRec(F).Handle)
end;

function CrtHasEvents(var F: Text): Boolean;
begin
	Result := CrtHasEvents(TTextRec(F).Handle)
end;

function CrtHasEvents(Input: THandle): Boolean;
var
	E: DWORD;
begin
	GetNumberOfConsoleInputEvents(Input, E);
	Result := E > 0
end;

function CrtGetEvent(var F: Text; var Event: INPUT_RECORD): Boolean;
begin
	Result := CrtGetEvent(TTextRec(F).Handle, Event)
end;

function CrtGetEvent(Input: THandle; var Event: INPUT_RECORD): Boolean;
var
	Readed: DWORD;
begin
	Result := ReadConsoleInput(Input, Event, 1, Readed) and (Readed > 0)
end;

{ Direct Access }

procedure CrtDirectWrite(var F: Text; X, Y: Integer; const Text: String; Attribute: Cardinal); overload;
begin
  CrtDirectWrite(TTextRec(F).Handle, X, Y, Text, Attribute);
end;

procedure CrtDirectWrite(Screen: THandle; X, Y: Integer; const Text: String; Attribute: Cardinal);
begin
	CrtDirectWrite(Screen, X, Y, PChar(Text), Length(Text), Attribute);
end;

procedure CrtDirectWrite(Screen: THandle; X, Y: Integer; P: PChar; L: Cardinal; Attribute: Cardinal);
var
	Buffer: array[0..0, 0..$ff] of CHAR_INFO;
	I: Integer;
	Size: COORD;
	R: SMALL_RECT;
begin
	L := L and $ff;

	for I := 0 to L - 1 do
	begin
		Buffer[0, I].AsciiChar := (P + I)^;
		Buffer[0, I].Attributes := Attribute;
	end;

	R.Left   := X;
	R.Top    := Y;
	R.Right  := X + Integer(L) - 1;
	R.Bottom := Y;

	Size.X := L;
	Size.Y := 1;

	WriteConsoleOutput(Screen, @Buffer, Size, ZeroCoord, R);
end;

{ Text Mode Graphics }

procedure CrtLine(var F: Text; X1, Y1, X2, Y2: Integer; const S: string; Attribute: Cardinal);
begin
	CrtLine(TTextRec(F).Handle, X1, Y1, X2, Y2, S, Attribute);
end;

procedure CrtLine(Screen: THandle; X1, Y1, X2, Y2: Integer; const S: string; Attribute: Cardinal);
var
	Step: TPoint;
	D, W, H, T: Integer;
	PX, PCX, PCY, PSX, PSY: PInteger;
begin
	Step.X := 1; if X1 > X2 then Step.X := -1;
	Step.Y := 1; if Y1 > Y2 then Step.Y := -1;

	if (X1 - X2) mod 2 <> 0 then Dec(X2, Step.X);
	Inc(Step.X, Step.X);

	W := Abs(X1 - X2) div 2;
	H := Abs(Y1 - Y2);

	if W >= H then
	begin
		PCX := @X1;
		PCY := @Y1;
		PSX := @Step.X;
		PSY := @Step.Y;
		PX := @X2;
	end
	else
	begin
		PCX := @Y1;
		PCY := @X1;
		PSX := @Step.Y;
		PSY := @Step.X;
		PX := @Y2;
		T := W; W := H; H := T;
	end;

	D := W div 2;
	repeat
		CrtDirectWrite(Screen, X1, Y1, S, Attribute);

		if PCX^ = PX^ then Break;

		Inc(D, H);
		if D >= W then
		begin
			Inc(PCY^, PSY^);
			Dec(D, W);
		end;
		Inc(PCX^, PSX^);
	until False;
end;

procedure CrtLine(var F: Text; X1, Y1, X2, Y2: Integer; const S: Char; Attribute: Cardinal);
begin
	CrtLine(TTextRec(F).Handle, X1, Y1, X2, Y2, S, Attribute)
end;

procedure CrtLine(Screen: THandle; X1, Y1, X2, Y2: Integer; const S: Char; Attribute: Cardinal);
var
	Step: TPoint;
	D, W, H, T: Integer;
	PX, PCX, PCY, PSX, PSY: PInteger;
begin
	Step.X := 1; if X1 > X2 then Step.X := -1;
	Step.Y := 1; if Y1 > Y2 then Step.Y := -1;

	W := Abs(X1 - X2) div 2;
	H := Abs(Y1 - Y2);

	if W >= H then
	begin
		PCX := @X1;
		PCY := @Y1;
		PSX := @Step.X;
		PSY := @Step.Y;
		PX := @X2;
	end
	else
	begin
		PCX := @Y1;
		PCY := @X1;
		PSX := @Step.Y;
		PSY := @Step.X;
		PX := @Y2;
		T := W; W := H; H := T;
	end;

	D := W div 2;
	repeat
		CrtDirectWrite(Screen, X1, Y1, S, Attribute);

		if PCX^ = PX^ then Break;

		Inc(D, H);
		if D >= W then
		begin
			Inc(PCY^, PSY^);
			Dec(D, W);
		end;
		Inc(PCX^, PSX^);
	until False;
end;

{  }

procedure CrtWriteAttribute(Screen: THandle;
				const R: TRect; Attribute: Cardinal);
var
	X, Y, Len: Integer;
	Attrs: array[0..159] of Word;
	S: COORD;
	Ret: DWORD;
begin
	if R.Top > R.Bottom then Exit;

	Len:= R.Right - R.Left + 1;
	if Len <= 0 then Exit;
	if R.Left + Len > 80 then Len := 80 - R.Left;

	for X:= Len - 1 downto 0 do
		Attrs[X]:= Attribute;

	S.X:= R.Left;
	for Y:= R.Top to R.Bottom do
	begin
		S.Y:= Y;
		WriteConsoleOutputAttribute(Screen, @Attrs, Len, S, Ret);
	end;
end;

procedure CrtWriteDirectTransparent(Screen: THandle;
				X, Y: Integer; const Text: String; ForeAttribute: Cardinal);
var
	Buffer: array[0..0, 0..159] of CHAR_INFO;
	L, I: Integer;
	Size: COORD;
	R: SMALL_RECT;
begin
	L := Length(Text);
	Assert(L <= 160);

	R.Left := X; R.Top := Y; R.Right := X + L - 1; R.Bottom := Y;
	Size.X := L; Size.Y := 1;

	ReadConsoleOutput(Screen, @Buffer, Size, ZeroCoord, R);

	for I := 0 to L - 1 do
	begin
		Buffer[0, I].AsciiChar := (PChar(Text) + I)^;
		Buffer[0, I].Attributes := Buffer[0, I].Attributes and $f0 or ForeAttribute;
	end;

	WriteConsoleOutput(Screen, @Buffer, Size, ZeroCoord, R);
end;

procedure PushAnyKey;
const
	Msg = '[ Please press any key to terminate ]';
var
	Input, Output: THandle;
	Done: DWORD;
	Cr: CONSOLE_CURSOR_INFO;
	InpRec: INPUT_RECORD;
begin
	Output := GetStdHandle(STD_OUTPUT_HANDLE);
	SetConsoleActiveScreenBuffer(Output);
	Cr.dwSize := 1;
	Cr.bVisible := False;
	SetConsoleCursorInfo(Output, Cr);

	SetConsoleTextAttribute(Output, 7 or BACKGROUND_RED);
	WriteConsole(Output, PChar(Msg), Length(Msg), Done, nil);

	Input := GetStdHandle(STD_INPUT_HANDLE);
	SetConsoleMode(Input, 0);

	FlushConsoleInputBuffer(Input);
	repeat
		ReadConsoleInput(Input, InpRec, 1, Done);
	until (Done = 1) and (InpRec.EventType = KEY_EVENT) and (InpRec.Event.KeyEvent.bKeyDown);
end;

initialization

	Rewrite(Output);
  Reset(Input);

finalization

	if (DebugHook <> 0) and IsCrt then PushAnyKey

end.


